// ----------------------------------------------------------------------------
//  FILE: imstrmbl.inl
//
//  This file contains the streaming macros.
//
//  COPYRIGHT:
//   IBM Open Class Library
//   (C) Copyright International Business Machines Corporation,  1997
//
//   Licensed Material - Program-Property of IBM - All Rights Reserved.
//
// Revision: 47 1.22.1.5 source/core/streams/imstrmbl.inl, oss, ioc.v400, trackPseudoLevel 
// ----------------------------------------------------------------------------

#ifndef _IMSTRMBL_INL_
#define _IMSTRMBL_INL_

#include <ilanglvl.hpp>
#include <iprimtyp.hpp>
#include <idatstrm.hpp>
#include <ityperep.hpp>

#if IC_RTTI_ENABLED

#define IMSTREAMABLE    virtual IMStreamable

#include <typeinfo.h>

#else // IC_RTTI_ENABLED

#define IMSTREAMABLE    IMStreamable

#endif // IC_RTTI_ENABLED

//
// external class declarations.
//

#if __IBMCPP__ >= 400
#pragma namemangling(compat)
#endif

class IStreamModule;
class IStreamIO;
class IMRefCounted;
class IVoidMasterPointer;

#pragma pack(push,4)
#pragma enum(4)

//
// IStreamableFactory
//
class IStreamableFactory
{
public:
        const ITypeRepresentation& typeRepresentation() const
    {
        return fTypeRep;
    };

protected:
        IStreamableFactory(IStreamModule* module);

friend class IStreamableGlobalTemplateHelper;
        virtual IMStreamable* createUninitialized(IMRefCounted*& refCounted) const = 0;
        virtual IMStreamable* createCopy(const IMStreamable& other) const = 0;
        virtual const char*     name() const = 0;

        IStreamableFactory();
        ~IStreamableFactory();
        IStreamableFactory(const IStreamableFactory& other);
        IStreamableFactory&     operator=(const IStreamableFactory& other);

private:
        ITypeRepresentation     fTypeRep;
        IStreamModule*  fModule;
        ITypeName*      fTypeName;

    friend class ITypeRepresentation;
        const ITypeName&        typeName();

    friend class IStreamFinder;
        static  const IStreamableFactory&       findFactory(const ITypeRepresentation& typeRepresentation);
};

#pragma enum(pop)
#pragma pack(pop)

inline void
streamableHelperAssignRefCounted(IMRefCounted*& refCounted, void*)
{
        refCounted = 0;
}

inline void
streamableHelperAssignRefCounted(IMRefCounted*& refCounted, IMRefCounted* temp)
{
        refCounted = temp;
}

#pragma pack(push,4)
#pragma enum(4)

template <class AType>
class IStreamableHelperFor : public IStreamableFactory
{
public:
#if     IC_RTTI_ENABLED
        IStreamableHelperFor(IStreamModule* module, ...)
                : IStreamableFactory(module)
                , fClassName(typeid(AType).name())
#else
        IStreamableHelperFor(IStreamModule* module, const char* className)
                : IStreamableFactory(module)
                , fClassName(className)
#endif
        {
        };
        virtual IMStreamable*   createCopy(const IMStreamable& other) const
        {
#if     IC_RTTI_ENABLED
#ifdef _MSC_VER // a Microsoft Bug
                return new AType(dynamic_cast<const AType&>(const_cast<IMStreamable&>(other)));
#else
                return new AType(dynamic_cast<const AType&>(other));
#endif
#else // IC_RTTI_ENABLED
                return new AType(*((const AType*)&other));
#endif // IC_RTTI_ENABLED
        };
        virtual IMStreamable*   createUninitialized(IMRefCounted*& refCounted) const
        {
                AType* temp = new AType();
                streamableHelperAssignRefCounted(refCounted, temp);
                return temp;
        };
        virtual const char*     name() const { return fClassName; };

private:
        const char*     fClassName;

        IStreamableHelperFor() {};
        IStreamableHelperFor(const IStreamableHelperFor<AType>& other) {};
        IStreamableHelperFor<AType>&    operator=(const IStreamableHelperFor<AType>& other) { return *this; };
};

#pragma enum(pop)
#pragma pack(pop)

#define IBEFRIENDSTREAMS(myclass) \
friend IMStreamable* IStreamableHelperFor<myclass>::createUninitialized(IMRefCounted*& refCounted) const;

#define StreamableDeclarationsMacro(className) \
public: \
        virtual const ITypeRepresentation&      typeRepresentation() const { return fgStreamableHelper.typeRepresentation(); }; \
        static const ITypeRepresentation&       staticTypeRepresentation() { return fgStreamableHelper.typeRepresentation(); }; \
private: \
        static  IStreamableHelperFor<className> fgStreamableHelper

#define StreamableTemplateDeclarationsMacro \
public: \
        virtual const ITypeRepresentation&      typeRepresentation() const { return fgStreamableHelper.typeRepresentation(); }; \
        static const ITypeRepresentation&       staticTypeRepresentation() { return fgStreamableHelper.typeRepresentation(); }

#if     IC_RTTI_ENABLED

#define StreamableDefinitionsMacro(className, moduleInstance) \
        IStreamableHelperFor<className> className::fgStreamableHelper(&moduleInstance)

#else // IC_RTTI_ENABLED

#define StreamableDefinitionsMacro(className, moduleInstance) \
        IStreamableHelperFor<className> className::fgStreamableHelper(&moduleInstance, #className)

#endif // IC_RTTI_ENABLED

class writeObjectInvalidForNonStreamableObjects;
class writeObjectInvalidForRefCountedObjects;
class writeAliasedObjectInvalidForNonStreamableObjects;
class writeAliasedObjectInvalidForNonRefCountedObjects;
class readObjectInvalidForNonStreamableObjects;
class readObjectInvalidForRefCountedObjects;
class readAliasedObjectInvalidForNonStreamableObjects;
class readAliasedObjectInvalidForNonRefCountedObjects;

#pragma pack(push,4)
#pragma enum(4)

class IStreamableGlobalTemplateHelper
{
public:
        //
        // write an object.  if typeRepresentation is given and is the type
        // of the object, the type of the object is not written to the stream.
        //
        static  void    writeUnaliasedObject(
                                                                        IStreamIO& streamIO,
                                                                        const IMStreamable* source);
        static  void    writeUnaliasedObject(
                                                                        IStreamIO& streamIO,
                                                                        const IMStreamable*     source,
                                                                        const ITypeRepresentation& typeRepresentation);

        //
        // write an aliased object.  if typeRepresentation is given and is the type
        // of the object, the type of the object is not written to the stream.
        //
        static  void    writeAliasedObject(
                                                                        IStreamIO& streamIO,
                                                                        const IMStreamable* source,
                                                                        const IMRefCounted* refCounted);
        static  void    writeAliasedObject(
                                                                        IStreamIO& streamIO,
                                                                        const IMStreamable*     source,
                                                                        const IMRefCounted* refCounted,
                                                                        const ITypeRepresentation& typeRepresentation);
        static  void    writeAliasedObject(
                                                                        IStreamIO& streamIO,
                                                                        const IMStreamable*     source,
                                                                        const IMRefCounted* refCounted,
                                                                        IBool isWrapped);

        //
        // read an object.  if typeRepresentation is not given, the stream must contain the
        // type information needed to create the object, which is then streamed into.
        // if typeRepresentation is given and the object on the stream is non-nil,
        // the typeRepresentation is used to create the object that is
        // streamed into.  the finder is used to find the factory used to create the
        // uninitialized object.  the default finder is used if the finder is not given.
        //
        static  IMStreamable*   readUnaliasedObject(IStreamIO& streamIO);
        static  IMStreamable*   readUnaliasedObject(
                                                                                                        IStreamIO& streamIO,
                                                                                                        const ITypeRepresentation& typeRepresentation);
        static  IMStreamable*   readUnaliasedObject(
                                                                                                        IStreamIO& streamIO,
                                                                                                        const IStreamFinder& finder);
        static  IMStreamable*   readUnaliasedObject(
                                                                                                        IStreamIO& streamIO,
                                                                                                        const IStreamFinder& finder,
                                                                                                        const ITypeRepresentation& typeRepresentation);

        //
        // read an aliased object.  if typeRepresentation is not given, the stream must
        // contain the type information needed to create the object, which is then streamed
        // into.  if typeRepresentation is given and the object on the stream is non-nil and
        // not aliased, the typeRepresentation is used to create the object that is
        // streamed into.  the finder is used to find the factory used to create the
        // uninitialized object.  the default finder is used if the finder is not given.
        //
        static  IMStreamable*   readAliasedObject(IStreamIO& streamIO);
        static  IMStreamable*   readAliasedObject(
                                                                                                        IStreamIO& streamIO,
                                                                                                        const ITypeRepresentation& typeRepresentation);
        static  IMStreamable*   readAliasedObject(
                                                                                                        IStreamIO& streamIO,
                                                                                                        const IStreamFinder& finder);
        static  IMStreamable*   readAliasedObject(
                                                                                                        IStreamIO& streamIO,
                                                                                                        const IStreamFinder& finder,
                                                                                                        const ITypeRepresentation& typeRepresentation);
        static  IMStreamable*   readAliasedObject(
                                                                                                        IStreamIO& streamIO,
                                                                                                        IVoidMasterPointer*& result,
                                                                                                        IBool isWrapped);

        //
        // create a polymorphic copy of a streamable object.
        //
        static  IMStreamable*   createCopy(const IMStreamable& source);
};

#pragma enum(pop)
#pragma pack(pop)

#if __IBMCPP__ >= 400
#pragma namemangling()
#endif

inline writeObjectInvalidForNonStreamableObjects&
writeObjectCheck(IDataStream& theStream, void* /*typedNil*/, void* /*typedNil*/)
{
        return *(writeObjectInvalidForNonStreamableObjects*)0;
}

inline IDataStream&
writeObjectCheck(IDataStream& theStream, IMStreamable* /*typedNil*/, void* /*typedNil*/)
{
        return theStream;
}

inline writeObjectInvalidForRefCountedObjects&
writeObjectCheck(IDataStream& theStream, IMStreamable* /*typedNil*/, IMRefCounted* /*typedNil*/)
{
        return *(writeObjectInvalidForRefCountedObjects*)0;
}

template <class AType>
inline void writeObject(const AType* theObject, IDataStream& theStream)
{
        IStreamableGlobalTemplateHelper::writeUnaliasedObject(*::writeObjectCheck(theStream, (AType*)0, (AType*)0).streamIO(), theObject);
}

template <class AType>
inline void
writeObject(const AType* theObject, const ITypeRepresentation& typeRep, IDataStream& theStream)
{
        IStreamableGlobalTemplateHelper::writeUnaliasedObject(*::writeObjectCheck(theStream, (AType*)0, (AType*)0).streamIO(), theObject, typeRep);
}

inline writeAliasedObjectInvalidForNonStreamableObjects&
writeAliasedObjectCheck(IDataStream& theStream, void* /*typedNil*/, void* /*typedNil*/)
{
        return *(writeAliasedObjectInvalidForNonStreamableObjects*)0;
}

inline writeAliasedObjectInvalidForNonRefCountedObjects&
writeAliasedObjectCheck(IDataStream& theStream, IMStreamable* /*typedNil*/, void* /*typedNil*/)
{
        return *(writeAliasedObjectInvalidForNonRefCountedObjects*)0;
}

inline IDataStream&
writeAliasedObjectCheck(IDataStream& theStream, IMStreamable* /*typedNil*/, IMRefCounted* /*typedNil*/)
{
        return theStream;
}

template <class AType>
inline void writeAliasedObject(const AType* theObject, IDataStream& theStream)
{
        IStreamableGlobalTemplateHelper::writeAliasedObject(*::writeAliasedObjectCheck(theStream, (AType*)0, (AType*)0).streamIO(), theObject, theObject);
}

template <class AType>
inline void
writeAliasedObject(const AType* theObject, const ITypeRepresentation& typeRep, IDataStream& theStream)
{
        IStreamableGlobalTemplateHelper::writeAliasedObject(*::writeAliasedObjectCheck(theStream, (AType*)0, (AType*)0).streamIO(), theObject, theObject, typeRep);
}

template <class AType>
inline void
writeCountedObject(const AType* theObject, IMRefCounted* refCounted, IBool isWrapped, IDataStream& theStream)
{
        IStreamableGlobalTemplateHelper::writeAliasedObject( *theStream.streamIO(), theObject, refCounted, isWrapped );
}

inline readObjectInvalidForNonStreamableObjects&
readObjectCheck(IDataStream& theStream, void* /*typedNil*/, void* /*typedNil*/)
{
        return *(readObjectInvalidForNonStreamableObjects*)0;
}

inline IDataStream&
readObjectCheck(IDataStream& theStream, IMStreamable* /*typedNil*/, void* /*typedNil*/)
{
        return theStream;
}

inline readObjectInvalidForRefCountedObjects&
readObjectCheck(IDataStream& theStream, IMStreamable* /*typedNil*/, IMRefCounted* /*typedNil*/)
{
        return *(readObjectInvalidForRefCountedObjects*)0;
}

template <class AType>
inline void readObject(AType*& theResult, IDataStream& theStream)
{
        IMStreamable* theStreamable = IStreamableGlobalTemplateHelper::readUnaliasedObject(*::readObjectCheck(theStream, (AType*)0, (AType*)0).streamIO());
#if     IC_RTTI_ENABLED
        theResult = dynamic_cast<AType*>(theStreamable);
#else
        theResult = (AType*)theStreamable;
#endif
}

template <class AType>
inline void
readObject(AType*& theResult, const ITypeRepresentation& typeRep, IDataStream& theStream)
{
        IMStreamable* theStreamable = IStreamableGlobalTemplateHelper::readUnaliasedObject(*::readObjectCheck(theStream, (AType*)0, (AType*)0).streamIO(), typeRep);
#if     IC_RTTI_ENABLED
        theResult = dynamic_cast<AType*>(theStreamable);
#else
        theResult = (AType*)theStreamable;
#endif
}

template <class AType>
inline void
readObject(AType*& theResult, const IStreamFinder& finder, IDataStream& theStream)
{
        IMStreamable* theStreamable = IStreamableGlobalTemplateHelper::readUnaliasedObject(*::readObjectCheck(theStream, (AType*)0, (AType*)0).streamIO(), finder);
#if     IC_RTTI_ENABLED
        theResult = dynamic_cast<AType*>(theStreamable);
#else
        theResult = (AType*)theStreamable;
#endif
}

template <class AType>
inline void
readObject(AType*& theResult, const IStreamFinder& finder, const ITypeRepresentation& typeRep, IDataStream& theStream)
{
        IMStreamable* theStreamable = IStreamableGlobalTemplateHelper::readUnaliasedObject(*::readObjectCheck(theStream, (AType*)0, (AType*)0).streamIO(), finder, typeRep);
#if     IC_RTTI_ENABLED
        theResult = dynamic_cast<AType*>(theStreamable);
#else
        theResult = (AType*)theStreamable;
#endif
}

inline readAliasedObjectInvalidForNonStreamableObjects&
readAliasedObjectCheck(IDataStream& theStream, void* /*typedNil*/, void* /*typedNil*/)
{
        return *(readAliasedObjectInvalidForNonStreamableObjects*)0;
}

inline readAliasedObjectInvalidForNonRefCountedObjects&
readAliasedObjectCheck(IDataStream& theStream, IMStreamable* /*typedNil*/, void* /*typedNil*/)
{
        return *(readAliasedObjectInvalidForNonRefCountedObjects*)0;
}

inline IDataStream&
readAliasedObjectCheck(IDataStream& theStream, IMStreamable* /*typedNil*/, IMRefCounted* /*typedNil*/)
{
        return theStream;
}

template <class AType>
inline void readAliasedObject(AType*& theResult, IDataStream& theStream)
{
        IMStreamable* theStreamable = IStreamableGlobalTemplateHelper::readAliasedObject(*::readAliasedObjectCheck(theStream, (AType*)0, (AType*)0).streamIO());
#if     IC_RTTI_ENABLED
        theResult = dynamic_cast<AType*>(theStreamable);
#else
        theResult = (AType*)theStreamable;
#endif
}

template <class AType>
inline void
readAliasedObject(AType*& theResult, const ITypeRepresentation& typeRep, IDataStream& theStream)
{
        IMStreamable* theStreamable = IStreamableGlobalTemplateHelper::readAliasedObject(*::readAliasedObjectCheck(theStream, (AType*)0, (AType*)0).streamIO(), typeRep);
#if     IC_RTTI_ENABLED
        theResult = dynamic_cast<AType*>(theStreamable);
#else
        theResult = (AType*)theStreamable;
#endif
}

template <class AType>
inline void
readAliasedObject(AType*& theResult, const IStreamFinder& finder, IDataStream& theStream)
{
        IMStreamable* theStreamable = IStreamableGlobalTemplateHelper::readAliasedObject(*::readAliasedObjectCheck(theStream, (AType*)0, (AType*)0).streamIO(), finder);
#if     IC_RTTI_ENABLED
        theResult = dynamic_cast<AType*>(theStreamable);
#else
        theResult = (AType*)theStreamable;
#endif
}

template <class AType>
inline void
readAliasedObject(AType*& theResult, const IStreamFinder& finder, const ITypeRepresentation& typeRep, IDataStream& theStream)
{
        IMStreamable* theStreamable = IStreamableGlobalTemplateHelper::readAliasedObject(*::readAliasedObjectCheck(theStream, (AType*)0, (AType*)0).streamIO(), finder, typeRep);
#if     IC_RTTI_ENABLED
        theResult = dynamic_cast<AType*>(theStreamable);
#else
        theResult = (AType*)theStreamable;
#endif
}

template <class AType>
inline void
readCountedObject(AType*& theResult, IVoidMasterPointer*& indirect, IBool isWrapped, IDataStream& theStream)
{
        IMStreamable* theStreamable = IStreamableGlobalTemplateHelper::readAliasedObject(*theStream.streamIO(), indirect, isWrapped);
#if     IC_RTTI_ENABLED
        theResult = dynamic_cast<AType*>(theStreamable);
#else
        theResult = (AType*)theStreamable;
#endif
}

template <class AType>
inline AType* copy(const AType& source)
{
        IMStreamable* theCopy = IStreamableGlobalTemplateHelper::createCopy(source);
#if     IC_RTTI_ENABLED
        return dynamic_cast<AType*>(theCopy);
#else
        return (AType*)theCopy;
#endif
}

template <class AType>
inline AType* copyPointer(const AType* source)
{
        if (source == 0)
                return 0;
        return ::copy(*source);
}

#endif
